#CHE210D Final Project, Spring 2012
#Kate Barteau

#import python modules
import numpy as np
import time

#import compiled fortran library
import chargewall
import atomwrite
import sys
import os

#whether or not to use the 3d visualization
UseVisual = True
if UseVisual:
    #import custom visualization library
    import atomvis


#========== GLOBAL VARIABLES ==========
#distance cutoff for pairwise interactions
#Cut = 2.5
#set the random number seed; useful for debugging
#np.random.seed(342324)

starttime = time.time()
#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.

#def CalcEnergy(Pos, PosChargeWall, FreeCharge):
def CalcEnergy(Pos, PosChargeWall, FreeCharge):    
    """Returns total potential energy calculated in fortran routine
Input:
    Pos: starting positions, (N,3) array
Output:
    PEnergy: value of potential energy, float
"""
    #global strength, screen, M, L, Cut
    global M, L, Cut, screen, strength
    #return charge.calcenergy(Pos, PosChargeWall, FreeCharge, strength, screen, M, L, Cut)
    return chargewall.calcenergy(Pos, PosChargeWall, FreeCharge, screen, strength, M, L, Cut)
    
    
#def CalcEnergyForces(Pos, PosChargeWall, FreeCharge, Forces):
def CalcEnergyForces(Pos, PosChargeWall, FreeCharge, Forces):
    """Returns total potential energy and forces on each atom as calculated
        in fortran routine
Input:
    Pos: starting postitions, (N,3) array
    Forces: forces on each atom, (N,3) array
Output:
    PEnergy: value of potential energy, float
    Forces: forces on each atom, (N,3) array (modified by routine)
"""
    #global strength, screen, M, L, Cut
    global M, L, Cut, screen, strength
    #return charge.calcenergyforces(Pos, PosChargeWall, FreeCharge, strength, screen, M, L, Cut, Forces)
    return chargewall.calcenergyforces(Pos, PosChargeWall, FreeCharge, screen, strength, M, L, Cut, Forces)
    
def VVIntegrate(Pos, Vel, Accel, PosChargeWall, FreeCharge):
    """Returns total potential energy and forces on each atom as calculated
        in fortran routine
Input:
    Pos: starting postitions, (N,3) array
    Forces: forces on each atom, (N,3) array
Output:
    Pos: starting positions, (N,3) array
    Vel: (N,3) array of atomic velocities
    Accel: acceleration factor
    KE
    PE
"""
    #global strength, screen, M, L, Cut
    global M, L, Cut, dt, screen, strength
    #return charge.calcenergyforces(Pos, PosChargeWall, FreeCharge, strength, screen, M, L, Cut, Forces)
    return chargewall.vvintegrate(Pos, Vel, Accel, PosChargeWall, FreeCharge, screen, strength, M, L, Cut, dt)
     
   
    
def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #identify global variables
    global M, L, Cut
    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [CalcEnergy(Pos + NormDir * x, PosChargeWall, FreeCharge) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = CalcEnergy(Pos + NormDir * dx, PosChargeWall, FreeCharge)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [CalcEnergy(Pos + NormDir * 2. * dx, PosChargeWall, FreeCharge)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [CalcEnergy(Pos + NormDir * Dists[-1], PosChargeWall, FreeCharge)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
	    Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = CalcEnergy(Pos + NormDir * d3, PosChargeWall, FreeCharge)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = CalcEnergy(Pos + NormDir * d3, PosChargeWall, FreeCharge)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosWall)            
            atomvis.Update(PosMin)
           
        else:
            #initialize the visualization window
            atomvis.Init(PosWall)            
            atomvis.Init(PosMin)
        
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    #identify global variables
    global M, L, Cut
    #initial search direction
    Forces = np.zeros_like(Pos)
    print("zero forces")
    PE, Forces = CalcEnergyForces(Pos, PosChargeWall, FreeCharge, Forces)
    print("new forces calc")
    Dir = Forces
    OldPE = 1.e300
    #iterative line searches
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        #print("a")
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = CalcEnergyForces(Pos, PosChargeWall, FreeCharge, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos


def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
    
    return Pos
    
def InitWall(L):
    """places 10 x 10 grid of atoms on lower z boundary to form wall
Input:
    L: box length
Output:
    Poswall: (Nwall^2,3) array of positions of charged lower surface
"""    
    Nwall = 10
    wallspac = L/10
    
#    dummywall = np.zeros((Nwall**2,4), float)
    PosWall = np.zeros((Nwall**2,3), float)
    
    r = wallspac * np.arange(Nwall, dtype=float) - 0.5*L
    walledge =-0.5*L
    print "lower surface is at ", walledge
    
    i=0
    for x in r:
        for y in r:
            PosWall[i] = np.array([x,y,walledge],float)
#            dummywall[i] = np.array([x,y,-walledge,0],float)
            i+=1
            
            if i >= Nwall**2:
                    return PosWall #, dummywall
    print PosWall
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosWall)
        else:
            #initialize the visualization window
            atomvis.Init(PosWall)
    return PosWall 

def InitIons(N, L, Pos, PosWall, polychargefrac, surfacechargefrac):
    """Distributes charge over polymer, distributes charge over wall, and adds counterions
Input:
    N,L
    Pos: N,3 array of polymer atom positions
    polychargefrac
    surfacechargefrac
Output:
    Pos: polymer atom positions AND ion atom positions  
    FreeCharge: charge on polymer atoms and counterions
    PosChargWall: positions of charged counterions in wlal
    SurfaceCharge: charge labelling of wall sites
    Ntot: number of unbound counterions and polymer atoms
"""
#initialize charge labelling array
    #Ntot = N + countercharge  
    PolyCharge = np.zeros((N,1),int)
   # 
   
#first evenly distribute charges on polymer   
    for i in range(1,N+1):
        if (i % polychargefrac) == 0: #repeat unit has charge
            PolyCharge[i-1] = -1
    totalcharge = -np.sum(PolyCharge)
    print "total charge on polymer is ", -totalcharge
    
    wallcharge = np.int(np.ceil(surfacechargefrac*totalcharge))
    countercharge = totalcharge - wallcharge
    print "ions in wall are ", wallcharge
    print "free ions are ", countercharge 

    #place free ions, add to positions array
    PosIon = np.zeros((countercharge,3),float)
    PosIon = (L*np.random.rand(countercharge,3) - 0.5*L)   
    Pos = np.vstack((Pos, PosIon))
    Ntot = len(Pos)
   
    #label free ions, add to labels array
    CounterCharge = np.ones((countercharge,1),int)
    FreeCharge = np.vstack((PolyCharge,CounterCharge))
   
 #randomly select atoms in wall to be charged, label 
    SurfaceCharge = np.zeros((100,1),int)   
    PosChargeWall = np.zeros((wallcharge,3),float)
    j=0   
    while j < wallcharge:
        k = np.random.randint(100)
        if (SurfaceCharge[k] == 1):
            j=j
        else:
            SurfaceCharge[k] = 1
            PosChargeWall[j] = PosWall[k]
            j+=1
    
    print "wall charge intialized to be ", np.sum(SurfaceCharge)

   # print SurfaceCharge
    SurfaceCharge = SurfaceCharge + 2
   # print SurfaceCharge
    return Pos, FreeCharge, PosChargeWall, SurfaceCharge, Ntot  
    
  

def RescaleVelocities(Vel, T):
    """Rescales velocities in th'import sitecustomize' failed; use -v for tracebacke system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel)
    #find velocity scale factor from ratios of kinetic energy
    VScale = np.sqrt(1.5 * len(Vel) * T / KE)
    Vel = Vel * VScale
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    return Vel


def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #global variables
    global M, L, Cut
    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Accel = CalcEnergyForces(Pos, PosChargeWall, FreeCharge, Forces)
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))

def RadiusofGyration(Pos,N):
    """calculate average distribution and average radius of gyration (squared)
    
"""
    global M, L
    Rg = []
    numpoly = N/M
    
    for i in range(0,numpoly):
        Posi = []
        low = i*M
        high = (i+1)*M
        Posi = Pos[low:high]
        Posi = Posi - L*np.rint(Posi/L)

        Rgi = 0.0
        rbar = Posi.mean(axis=0)

        for j in range(0,M):
            rj = Posi[j,:]
            rj = rj - rbar
            
            Rgi += np.dot(rj,rj)
        Rgi = Rgi/M
        #g.write("%8.6f\t" %(Rgi))
        Rg.append(Rgi)
                 
    return Rg 

def ZHistogram(Pos, FreeCharge, L, Nbin):
	"""Returns 4 histograms as function of z position
		1) all free beads
		2) polymer density
          	3) charged polymer density
		4) counter ions only

Input:
	Pos(Ntot,3) array of polymer and counterion positions
	FreeCharge(Ntot,1) list of charges corresponding to Pos
	L: Simulation box size
	Nbin: number of bins

Output:
	zall: [Nbin,1] list of counts of atoms 
	zpoly: [Nbin,1] list of counts of polymer atoms 
	zchargepoly: [Nbin,1] list of counts of charged polymer atoms 
	zions: [Nbin,1] list of counts of free counterions 
"""

	
	binwidth = 1.0*L/Nbin
	
	binval = np.zeros((Nbin+1,1),float)
	zall = np.zeros((Nbin,1),int)
	zpoly = np.zeros((Nbin,1),int)
	zchargepoly = np.zeros((Nbin,1),int)
	zions = np.zeros((Nbin,1),int)

	for i in range(0,Nbin+1):
		binval[i] = i*binwidth - L/2. #set lower value of bin

    #iterate through positions
	for k in range(0, len(Pos)):#look at every position
		#cycle through bins
             j=0
             while (j < Nbin):
                    if (Pos[k,2] >= binval[j] and Pos[k,2] < binval[j+1]):
                        zall[j] +=1
                        
                        if (FreeCharge[k] != 1) :
                            zpoly[j]+=1
                        if(FreeCharge[k] == -1):
                            zchargepoly[j]+=1
                        if(FreeCharge[k] ==1):
                            zions[j]+=1
                    j+=1
                           	

	return binval, zall, zpoly, zchargepoly, zions         

#### YOUR CODE HERE ####

#### YOU WILL NEED TO DEFINE THE FOLLOWING GLOBAL ####
#### VARIABLES FOR THE CODE ABOVE TO WORK         ####
M = 32
T = 1.0
dt = 0.001
f = file("video2.txt", "a")
#g = file("wallRAWevenRg.txt","a")
#h = file("histogramtest_wall_surfacefrac_all_polyfrac_allevenRg.txt", "a")

screen = 0.001
strength = 5
N = 64

polychargefrac = 4
surfacechargefrac = 0.8
NStepsEquil1 = 10000
NStepsEquil2 = 10000
NStepsProd = 100000
ProdFreq = 100


p = 0.001
L = (N/p)**(1./3.)
Cut = 2.5
RescaleFreq = 500


#Set up historams
Nbin = 100
numhist = 0
zall = np.zeros((Nbin,1),int)
zpoly = np.zeros((Nbin,1),int)
zchargepoly = np.zeros((Nbin,1),int)
zions = np.zeros((Nbin,1),int)
zavg = 0.

            
#initialize polymers, walls, and ions
Pos = InitPositions(N,L)
PosWall = InitWall(L)
Pos, FreeCharge, PosChargeWall, SurfaceCharge, Ntot = InitIons(N, L, Pos, PosWall, polychargefrac, surfacechargefrac)

#save label list for Chimera
atomlabel = np.vstack((FreeCharge[0:M],FreeCharge[N:Ntot], SurfaceCharge))
b=atomlabel.flatten()
Labels = ['C' if x==0 else 'O' if x==-1 else 'H' if x==1 else 'P' if x==2 else 'N' for x in b]
Pdb = atomwrite.pdbfile("file002.pdb", AtomNames=Labels) 


PE, Pos = ConjugateGradient(Pos, 0.001, 1.0e-8, 1.0e-10)
Vel = InitVelocities(Ntot,T)
Accel = InitAccel(Pos)


totalt = int(2./dt) 
prodsteps = int(5*totalt) 
Ener = [] #list of total energy
           

#print "first equilibration begun"
for i in range(0, NStepsEquil1):
               #if using visualization, update the display
     if UseVisual:
          if atomvis.Initialized:
              #update the positions
              atomvis.Update(Pos)
          else:
              #initialize the visualization window
              atomvis.Init(Pos)
     if i % RescaleFreq == 0:
                   
             Vel = RescaleVelocities(Vel,T)
             Pos, Vel, Accel, KE, PE = VVIntegrate(Pos, Vel, Accel, PosChargeWall, FreeCharge)

            #print "first equilibration complete"
            #print "second equilibration"
for i in range(0, NStepsEquil2):
   #if using visualization, update the display
     if UseVisual:
           if atomvis.Initialized:
                 #update the positions
               atomvis.Update(Pos)
           else:
                        #initialize the visualization window
               atomvis.Init(Pos)
     if i % RescaleFreq == 0:
                    # print "second equil", i
           Vel = RescaleVelocities(Vel,T)
           Pos, Vel, Accel, KE, PE = VVIntegrate(Pos, Vel, Accel, PosChargeWall, FreeCharge)
           TotalE = KE + PE
           Ener.append(TotalE)
           #print TotalE
Ener = np.array(Ener,float)
    
TargetT = 2*(Ener.mean()-PE)/(3*(N-1))
Vel = RescaleVelocities(Vel,TargetT)

Pos0 = Pos.copy()
print "second equilibration complete"

#create array to send to pdb file
Posvid = []            
#atomlabels = []

#production run
#g.write("\nRg for polycharge frac %i\t and surface charge frac%6.3f simulation %i" % (polychargefrac, surfacechargefrac,l+1))   
for i in range(0, NStepsProd):     
    Pos, Vel, Accel, KE, PE = VVIntegrate(Pos, Vel, Accel, PosChargeWall, FreeCharge)
    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(Pos)
        else:
            #initialize the visualization window
            atomvis.Init(Pos)
    if (i > 0 and i % ProdFreq == 0):
         print "production step ",i
         #print len(Pos)
         postest = np.zeros((100,3),float)
         postest[:,2]+=1
         #print PosWall
         poswalla = PosWall-2.0*postest - 40.*wallshift
        # print poswalla
         Posvid = np.vstack((Pos[0:M]-40.*wallshift[0:M], Pos[N:Ntot]-40.*wallshift[N:Ntot],poswalla))
         Pdb.write(Posvid)
         #print len(Posvid)
        
        binvali, zalli, zpolyi, zchargepolyi, zionsi = ZHistogram(Pos, FreeCharge, L, Nbin)
        zall+=zalli
        zpoly+=zpolyi
        zchargepoly+=zchargepolyi
        zions+=zionsi
        for x in range(0,N):
            zavg += Pos[x,2]
            g.write("\nproduction step %i\t" %(i))    
            Rg = RadiusofGyration(Pos,N)
            # print Rg
            Rg = np.array(Rg,float)
            # print Rg.mean()
            f.write("%i\t%6.3f\t%i\t%i\t%8.6f\n" % (polychargefrac, surfacechargefrac, l, i, Rg.mean()))
            numhist+=1
            
                            
print "production run complete"
       

zavg = zavg*ProdFreq/(N*NStepsProd*simulationnumber)
zallTot = np.sum(zall)
zpolyTot = np.sum(zpoly)
zchargepolyTot = np.sum(zchargepoly)
zionsTot = np.sum(zions)
#
#
h.write("histogram for %i polymer atoms with chain length %i density %6.3f, temperature %6.3f\n" %(N, M, p, T))
h.write("charge distribution of %i and sufrace fraction %6.3f, screening length %6.3f and ion strength %i" %(polychargefrac, surfacechargefrac, screen, strength))
h.write("average z position of polymers is %8.6f and %i histograms collected\n" %(zavg,numhist))
h.write("binval\tzall\tzpoly\tzchargepoly\tzions\n")
for i in range(0,Nbin):
    h.write("%8.5f\t%i\t%i\t%i\t%i\n" %(binvali[i],zall[i],zpoly[i],zchargepoly[i],zions[i]))
    h.write("Totals\t%i\t%i\t%i\t%i\n" %(zallTot,zpolyTot,zchargepolyTot,zionsTot))

f.close()
Pdb.close()
#g.close()
#h.close()
stoptime = time.time()
print"\n", stoptime-starttime, "seconds"
